require 'ap'
require 'pp'


# 	класс работает с образом
# 	new(filename) открыть файо
# 	get_cpm_block(block_num) прочитать блок с диска
#   get_dpb получить блок параметров диска
#   get_system получить образ системы (корвет специфик?)

class Abstract_CPM_Disk
	attr_reader :file_name
	attr_reader :dpb

	def initialize(file_name)
		@file_name=file_name
	end

	def get_cpm_block(block_num)
		raise "not implemented yet"
	end

	def get_dpb
		raise "not implemented yet"
	end

	def get_system
		raise "not implemented yet"
	end
end

class DPB
	attr_accessor :spt   		# No. of Logical ( 128-byte ) Sectors per Logical Track
	attr_accessor :bsh   		# Block Shift - Block Size is given by 128 * 2^(BSH)
	attr_accessor :blm   		# Block Mask - Block Size is given by 128 * (BLM +1)
	attr_accessor :exm   		# Extent Mask ( see opposite )
	attr_accessor :dsm   		# объем памяти на диске в блоках минус 1
	attr_accessor :drm   		# число входов в директорию диска минус 1
	attr_accessor :al0   		# определяет, какие блоки зарезервированы
	attr_accessor :al1   		# под директорию
	attr_accessor :cks   		# размер вектора контроля директории
	attr_accessor :ofs   		# число системных дорожек на диске


		#http://sharpmz.org/dpb.htm
		#http://sharpmz.org/succpminfo06.htm

		#BSH
		#Data allocation "Block Shift Factor"
		#Number of 128 bytes sectors per "Allocation Block" 
		#( determined by the data block allocation size. Stored as 2's-logarithm. 
		#See notes below ).

		#BLM
		#"Data allocation Block Mask"
		#To get the sector index of the Allocation Block the sector number is to 
		#combine logically by AND with this mask. (( 2^BSH)-1). See notes below.


		#The values of BSH and BLM implicitly determine the data allocation size, BLS,
		#which is not an entry in the DPB. 
		#Given that the designer has selected a value for BLS, 
		#the values of BSH and BLM are shown in the following table:

		#BLS 	BSH	BLM
		# 1,024	  3   7
		# 2,048	  4  15
		# 4,096   5  31
		# 8,192   6  63
		#16,384   7 127


		#EXM	"Extent Mask"
		#Number of extents per directory entry ( stored as 2's logarithm ). 
		#Determined by the data block allocation size and the number of disk blocks.

		#DSM	
		#Total storage capacity of the disk drive. 
		#Number of the last Allocation Block ( Number of entries per disk -1 )

		#DRM	
		#Number of the last directory entry ( Number of entries per disk -1 ). 
		#Total number of directory entries that can be stored on this drive.
		#( AL0, AL1 determine reserved directory blocks ).

		#AL0& AL1
		#
		#These parameters, taken BITWISE, indicate which of the first 16 CP/M blocks ( numbered 00 - 15 ) are allocated to the Directory. For example:-
		#$80 00 = 10000000 00000000 i.e. the Directory is allocated BLOCK 00 only
		#$C0 00= 11000000 00000000 i.e. the Directory is allocated BLOCKS 00 & 01

		#CKS "Check area Size"
		#Size of the directory check vector. 
		#Number of directory entries to check for disk change. Zero for a HDD.

		#OFF	
		#Number of system reserved tracks at the beginning of the ( logical ) disk
end

class KDI_dpb < DPB
	def initialize(encoded_params)
		@spt        ,		# No. of Logical ( 128-byte ) Sectors per Logical Track
		@bsh        ,		# Block Shift - Block Size is given by 128 * 2^(BSH)
		@blm        ,		# Block Mask - Block Size is given by 128 * (BLM +1)
		@exm        ,		# Extent Mask ( see opposite )
		@dsm        ,		# объем памяти на диске в блоках минус 1
		@drm        ,		# число входов в директорию диска минус 1
		@al0        ,		# определяет, какие блоки зарезервированы
		@al1        ,		# под директорию
		@cks        ,		# размер вектора контроля директории
		@ofs        = encoded_params.unpack("SCCCSSCCSS")
	end
end

class KDI_disk_params

		# ;Eфизические параметры дискаF

		#  1:    DW LoadAdress ; адрес, с которого начинается загрузка
		#  3:    DW RunAdress  ; адрес,куда передается управление после загрузки
		#  5:    DW Count      ; число загружаемых физических секторов
		#  7:    DB SizeDisk   ; если значение байта 1 -8" диск, если 0 - 5.25"
		#  8:    DB Density    ; способ записи: 0 - FM; 1 - MFM
		#  9:    DB TpI        ; число дорожек  на дюйм: 0 - 48 TpI 1 - 96 TpI, 2 - 135 TpI
		# 10:    DB SkewFactor ; если значение байта 1 - фактора смещения
		#                      ; нет. Если <> 1 - байты 33-128
		#                      ; информационного сектора содержат таблицу
		#                      ; перевода секторов, причем значение байта
		#                      ; равно размеру таблицы.
		# 11:    DB SecSize    ; размер физического сектора:0 - 128 байт, 1 - 256 байт, 2 - 512 байт и  3 - 1024
		# 12:    DB InSide     ; информация о сторонах диска: 
		#					   ; 0 - односторонний диск, 1 - двухсторонний,
		#                      ; причем сектора нумеруются от 1 до n 
		#                      ; с каждой стороны, четные логические
		#                      ; дорожки с нулевой стороны, нечетные с
		#                      ; первой.
		# 13:    DW SecPerTrack; число физических секторов на дорожке
		# 15:    DW TrkPerDisk ; число дорожек на диске (с одной стороны)
		# ;Eлогические параметры дискаF

		# 17:    DW SPT        ; число логических секторов на дорожке
		# 19:    DB BSH        ; фактор сдвига блока распределения данных
		# 20:    DB BLM        ; маска блока распределения данных
		# 21:    DB EXM        ; маска размера
		# 22:    DW DSM        ; объем памяти на диске в блоках минус 1
		# 24:    DW DRM        ; число входов в директорию диска минус 1
		# 26:    DB AL0        ; определяет, какие блоки зарезервированы
		# 27:    DB AL1        ; под директорию
		# 28:    DW CKS        ; размер вектора контроля директории
		# 30:    DW OFS        ; число системных дорожек на диске
		# 32:    DB CheckSum   ;контрольная сумма служебной информации
		#                                                      CS[1-31]+66H

	attr_accessor :LoadAdress 
	attr_accessor :RunAdress  
	attr_accessor :Count      
	attr_accessor :SizeDisk   
	attr_accessor :Density    
	attr_accessor :TpI        
	attr_accessor :SkewFactor 
	attr_accessor :SecSize    
	attr_accessor :InSide     
	attr_accessor :SecPerTrack
	attr_accessor :TrkPerDisk 

	def initialize(encoded_params)
		@LoadAdress ,
		@RunAdress  ,
		@Count      ,
		@SizeDisk   ,
		@Density    ,
		@TpI        ,
		@SkewFactor ,
		@SecSize    ,
		@InSide     ,
		@SecPerTrack,
		@TrkPerDisk =encoded_params.unpack("SSSCCCCCCSS")
	end
end	
	
class KDI < Abstract_CPM_Disk	
	def initialize(disk_image)
		super(disk_image)

		@kdi_raw_image=File.open(@file_name,"rb").read
		#@kdi_image=kdi_raw_image.split('').map { |e| e.ord  }

		raise "Too small image" unless @kdi_raw_image.size > 128
		raise "Invalid Sector0 CRC" unless is_kdi_correct?

		@kdi_params=KDI_disk_params.new(@kdi_raw_image[0..15])
		@dpb=KDI_dpb.new(@kdi_raw_image[16..30])

		@block0_offset=@kdi_params.Count*(0x80<<@kdi_params.SecSize)
		p [:block0_offset,@block0_offset]
	end

	def get_dpb
		@dpb
	end

	def get_cpm_block(block_num)
		block_size=128*(1<<@dpb.bsh)

		b0=@block0_offset+(block_size*block_num)
		b1=b0+block_size
		@kdi_raw_image[b0...b1]
	end

	def get_system

		if @kdi_params.Count>0
			sector_size=0x80<<@kdi_params.SecSize
			@kdi_raw_image[0...(@kdi_params.Count*sector_size)]
		else
			""
		end
	end

	private

	def is_kdi_correct?
		calculated_crc =@kdi_raw_image[0..31-1].split('').reduce(0x66){|sum,n| (sum+n.ord)&0xff}
		disk_crc       =@kdi_raw_image[31].ord
		calculated_crc == disk_crc
	end
	# def to_s
	# 	[@file_name,@dpb,@kdi_params]
	# end

end


if __FILE__ == $0 #PROGRAM_NAME
	a=KDI.new("./KDI/disk1.kdi")
	ap a.dpb
end